/* Altre funzioni C per le stringhe */

#include "string2.h"

#ifndef MIN
  #define MIN(a,b)         (((a)<=(b))?(a):(b))
  #define LOCAL_MIN
#endif

#ifndef MIN3
  #define MIN3(a,b,c)      (MIN(MIN((a),(b)), (c)))
  #define LOCAL_MIN3
#endif



/* Rovescia una stringa */
/* Tempo = O(strlen(str)) */
char *strrev(char *str) {
  int t, i, l = strlen(str);
  for (i=0; i<(l>>1); ++i) {
    t = str[i];
    str[i] = str[l-i-1];
    str[l-i-1] = t;
  }
  return str;
}

/* Converte la stringa str in minuscolo */
/* Tempo = O(strlen(str)) */
char *strlower(char *str) {
  char *p=str-1;
  while (*++p) *p = tolower(*p);
  return str;
}


/* Converte la stringa str in maiuscolo */
/* Tempo = O(strlen(str)) */
char *strupper(char *str) {
  char *p=str-1;
  while (*++p) *p = toupper(*p);
  return str;
}


/* come strstr, ma si limita ai primi lim byte di where. */
/* Tempo = O(min(lim,strlen(where))) */
char *strnstr(char *where, const char *what, int lim) {
  int p=0,psave,i;
  if (what[0] == 0) return where;
  if (lim == 0) return NULL;

  while (1) {
    while (where[p] != 0 && where[p] != what[0] && p < lim) p++;

    if (where[p] == 0 || p >= lim) break;
    i = 0, psave = p;
    while (what[i] != 0 && where[p] == what[i] && p < lim) p++, i++;
    if (what[i] == 0) return where+psave;
    if (p >= lim) break;
  }
  return NULL;
}


/* elimina il carattere spc all'inizio di str */
/* Tempo = O(strlen(str)) */
char *strlstrip(char *str, char spc) {
  int i = 0, p = 0;
  while (str[i] && str[i] == spc) i++;
  if (i) {
    str[p] = str[i];
    while (str[i]) str[++p] = str[++i];
  }
  return str;
}


/* elimina il carattere spc alla fine di str */
/* Tempo = O(strlen(str)) */
char *strrstrip(char *str, char spc) {
  int i = 0, lp = -1;
  while (str[i]) {
    if (str[i] == spc) {
      if (lp == -1) lp = i;
    } else
      lp = -1;
    i++;
  }
  if (lp != -1) str[lp] = 0;
  return str;
}


/* elimina il carattere spc all'inizio e alla fine di str */
/* Tempo = O(strlen(str)) */
char *strstrip(char *str, char spc) {
  int si, i = 0, p = 0, lp = -1;
  while (str[i] && str[i] == spc) i++;
  si = i;
  while (str[i]) {
    if (str[i] == spc) {
      if (lp == -1) lp = i;
    } else
      lp = -1;
    i++;
  }
  if (lp != -1) str[lp] = 0;
  str[p] = str[si];
  while (str[si]) str[++p] = str[++si];
  return str;
}


/* elimina qualsiasi carattere di search dall'inizio di str */
/* Tempo = O(strlen(str)*(strlen(search)+1)) */
char *strlsstrip(char *str, const char *search) {
  int i = 0, p = 0;
  while (str[i]) {
    int j = 0;
    while(search[j] && str[i] != search[j]) j++;
    if (!search[j]) break;
    ++i;
  }
  if (i) {
    str[p] = str[i];
    while (str[i]) str[++p] = str[++i];
  }
  return str;
}


/* elimina qualsiasi carattere di search dalla fine di str */
/* Tempo = O(strlen(str)*strlen(search)) */
char *strrsstrip(char *str, const char *search) {
  int i = 0, lp = -1;
  while (str[i]) {
    int j = 0;
    while(search[j] && str[i] != search[j]) j++;
    if (search[j]) {
      if (lp == -1) lp = i;
    } else
      lp = -1;
    ++i;
  }
  if (lp != -1) str[lp] = 0;
  return str;
}


/* elimina ognuno dei caratteri in search all'inizio e alla fine di str */
/* Tempo = O(strlen(str)*(strlen(search)+1)) */
char *strsstrip(char *str, const char *search) {
  int si, i = 0, p = 0, lp = -1;
  while (str[i]) {
    int j = 0;
    while(search[j] && str[i] != search[j]) j++;
    if (!search[j]) break;
    ++i;
  }
  si = i;
  while (str[i]) {
    int j = 0;
    while(search[j] && str[i] != search[j]) j++;
    if (search[j]) {
      if (lp == -1) lp = i;
    } else
      lp = -1;
    ++i;
  }
  if (lp != -1) str[lp] = 0;
  str[p] = str[si];
  while (str[si]) str[++p] = str[++si];
  return str;
}


/* elimina il carattere dc dalla stringa */
/* Tempo = O(strlen(str)) */
char *strdelchr(char *str, char dc) {
  int i = 0, j = 0;
  while (1) {
    while(str[j] && str[j] == dc) ++j;
    if (str[j])
      str[i++] = str[j++];
    else {
      str[i] = 0;
      break;
    }
  }
  return str;
}

/* elimina tutti i caratteri di ds stringa str */
/* Tempo = O(strlen(str)*strlen(ds)) */
char *strdelschr(char *str, const char *ds) {
  int i = 0, j = 0;
  while (1) {
    while(str[j]) {
      int k = 0;
      while(ds[k] && str[j] != ds[k]) k++;
      if (!ds[k]) break;
      ++j;
    }
    if (str[j])
      str[i++] = str[j++];
    else {
      str[i] = 0;
      break;
    }
  }
  return str;
}


/* sostituisce tutte le occorrenze di search con susbt */
/* Tempo = O(strlen(str)) */
char *strsubchr(char *str, char search, char subst) {
  int i = 0;
  for ( ; str[i]; i++) {
    if (str[i] == search) str[i] = subst;
  }
  return str;
}


/* sostituisce i caratteri in search con i rispettivi caratteri in subst.*/
/* search e subst devono avere la stessa lunghezza */
/* Tempo = O(strlen(str)*strlen(search)) */
char *strsubschr(char *str, const char *search, const char *subst) {
  int i = 0;
  for ( ; str[i]; i++) {
    int j = 0;
    while(search[j] && str[i] != search[j]) j++;
    if (search[j]) str[i] = subst[j];
  }
  return str;
}

/* calcola la distanza di Hamming delle 2 stringhe
   se non hanno la stessa lunghezza si suppone che la restante
   parte della maggiore sia tutta diversa */
/* Tempo = O(max(strlen(str1),strlen(str2))) */
int strhamdist(const char *str1, const char *str2) {
  int d = 0, i=0, j=0;
  const char *p;
  for ( ; str1[i] && str2[j]; i++, j++) {
    if (str1[i] != str2[j]) ++d;
  }
  if (str1[i]) p = str1+i; else p = str2+j;
  while (*p++) d++;
  return d;
}


/* Calcola la distanza di Levenshtein tra 2 stringhe.
 * Restituisce -1 se non c'è sufficiente memoria. */
/* Tempo = O(strlen(str1)*strlen(str2)) */
int strlevdist(const char *str1, const char *str2) {
  int i,j, *d, cost;
  int l1 = strlen(str1);
  int l2 = strlen(str2);

  if (l1 == 0 || l2 == 0) return l1+l2;
  ++l1; ++l2;

  d = (int *)malloc(l1*l2*sizeof(int)+8);
  if (d == NULL) return -1;

  for (j=0; j<l1; j++) *(d+j*l2) = j;
  for (i=1; i<l2; i++) *(d+i) = i;

  for (j=1; j<l1; j++) {
    for (i=1; i<l2; i++) {
      cost = (str1[j-1] != str2[i-1]);
      *(d+j*l2+i) = MIN3( *(d+(j-1)*l2+i)+1, *(d+j*l2+i-1)+1, *(d+(j-1)*l2+i-1)+cost );
    }
  }
  int r = *(d+l1*l2-1);
  free(d);
  return r;
}


/* calcola la lunghezza della nuova stringa ottenuta sostituendo search con susbt */
/* non molto efficiente */
/*
int strsubstr_calc(const char *str, const char *search, const char *subst) {
  char *p;
  int stimes = 0, 
      selen = strlen(search),
      sulen = strlen(subst);
  
  p = strstr(str, search);
  while (p) {
    stimes++;
    p = strstr(p+selen, search);
  }
  return strlen(str)+stimes*(sulen-selen);
}
*/

/* sostituisce search con subst in str
   se subst e' piu' lunga di search, si assume che ci sia abbastanza spazio in memoria */
/* */
/*
char *strsubstr(const char *str, const char *search, const char *subst) {
  char *p;
  int selen = strlen(search),
      sulen = strlen(subst);
  
  //if (sulen > selen) CONTROLLA SPAZIO!!;
  p = strstr(str, search);
  while (p) {
    
    //p = strstr(p+selen, search);
  }
  return str;
}
*/

#ifdef LOCAL_MIN3
  #undef LOCAL_MIN3
  #undef MIN3
#endif

#ifdef LOCAL_MIN
  #undef LOCAL_MIN
  #undef MIN
#endif

